#pragma once
#include "WMAllocator.h"
#include "Globals.h"

namespace WorldMachine {

	class WMCoreInterface;
	class PacketTypeMany;
	extern WMCoreInterface *core;

	typedef int TypeHandle;

/** 
	A class to help identify and query a given type. Ports use this; it allows for indication of acceptance of various types, including user-defined types.
*/
class PacketType {
public:
	PacketType();
	/// Create a type from a type handle. type should be a primitive defined by the WM core, or the handle to a UDT)
	PacketType(TypeHandle primitive_or_udt);

	/// Load and save a description of the packet TYPE (not the packet itself).
	/// For built-in types this is simply an integer code value; For user defined types, it is the full match-level descriptor of the type
	bool Load(std::istream &in);
	bool Save(std::ostream &out) const;

	bool operator==(PacketType const&other) const;
	bool operator!=(PacketType const&other) const;

	friend PacketTypeMany operator&(PacketType const&a, PacketType const&b);
	friend PacketTypeMany operator|(PacketType const&a, PacketType const&b);

	TypeHandle getHandle() const;
private:
	PacketType &operator^(PacketType const&other);
	PacketType &operator~();
	TypeHandle type;
};

//PacketTypeMany operator&(PacketType const&a, PacketType const&b);
PacketTypeMany operator|(PacketType const&a, PacketType const&b);
PacketTypeMany operator|(PacketTypeMany &a, PacketType const&b);

/**
Provide a way to collect multiple packet types indicate 

*/
class PacketTypeMany {
public:
	PacketTypeMany();
	PacketTypeMany(TypeHandle const type);
	PacketTypeMany(PacketType const type);
	PacketTypeMany(PacketTypeMany const&other);

	/// Convert a multiple-type back to a single type, by ignoring all types other than the first one.
	PacketType getPrimaryType() const;

	void addType(PacketType const&othertype);

	bool acceptsType(PacketTypeMany const&othertype) const;
	bool acceptsType(PacketType const&othertype) const;
	bool acceptsType(TypeHandle const othertype) const;

//	void Save( std::ostream & out );
private:
	static const int MAX_TYPES = 16;
	PacketType types[MAX_TYPES];
	int num_types;
};

class UserTypeHandle;

/**
	DefineUserType is the interface that defines user datatypes. These are roughly equivelent to a struc or class in World Machine's type system, that your devices can pass around and
	operate on.

	A DefineUserType is created by the factory provided in the WMGlobal interface.

	To use:

	1) Define a unique name, namespace, and the structure of the type using the fluent interface of DefineUserType as shown below
	2) Register your user datatype. All datatypes with the same name within the same namespace define the same type. You will receive a handle to the type.

	ie:

	auto myRGBADescription = DefineUserType("RGBA").in_namespace("WorldMachine").make_private().member("rgb", WMP_rgb).member("a", WMP_hfield).member("FurtherInsideInfo", InnerHandle);
	TypeHandle myRGBAType = WorldMachine::core->memory->DefineType(myRGBADescription);


	For convenience, a proxy has been provided for TypeHandle; see the documentation on that class for more details, but it is used simply, like:

	UserTypeHandle myTypeHandle = DefineUserType("RGBA").in_namespace("WorldMachine").make_private().member("rgb", WMP_rgb).member("a", WMP_hfield).member("FurtherInsideInfo", InnerHandle);

	the type handle can then be passed to any place expecting a type:

	in(0)->SetType(WMP_rgb);
	in(1)->SetType(WMP_hfield);
	out(0)->SetType(myRGBAType);

	Note that if you attempt to define the same type in two different places but with two different definitions, the second definition will throw an exception! They must match exactly
	to be allowed. You can use the namespace functionality for versioning in case you need to change how the user type is defined.
	
*/
class DefineUserType {
public:
// Fluent Interface members.
// Chain the named parameters together to create your type.
	DefineUserType(const DefineUserType &other);
	DefineUserType(const char *udt_name);

	/// specify that the type should be private (unable to have its internals inspected)
	DefineUserType &make_private();

	/// specify that the type should allow implicit conversions to and from compatible types. The default is to disallow
	/// (types that have the same number, type, and ordering of members, but not necessarily the same names)
	DefineUserType &allow_implicit_conversion();

	/// provide namespace for this type to allow versioning, and prevent inadvertent collisions between your defined types 
	DefineUserType &in_namespace(const char*nmspace);

	/// Add a simple type member to this UDT.
	/// TODO : We don't currently actually enforce the required param
	DefineUserType &member(const char *name, PacketType type, bool required = true);

	/// Add a user-defined type to this UDT
	/// TODO : We don't currently actually enforce the required param
	DefineUserType &member(const char *name, UserTypeHandle& type, bool required = true);

// Regular members

	bool Load(std::istream &in);		///< Load type descriptor from stream. Note that this WILL recursively load and register any user types that are members, although it will NOT register this type automatically.
	bool Save(std::ostream &out) const; ///< Save type descriptor to stream, also recursively saving any member user type descriptors

	bool operator==(DefineUserType const&other) const;
	bool operator!=(DefineUserType const&other) const;

	/// If this user type relies has another user type for a member, this instantiates them as well.
	void createDeferredMembers();

	/// Private types do not allow the end user to see the members of the type
	bool isPrivate() const;

	/// Retrieve name
	const char*getName() const;

	/// Retrieve namespace
	const char*getNamespace() const;

	/// Look up a member. Will throw an exception if the name or index is not present in the type
	TypeHandle getMember(const char*name) const;
	TypeHandle getMember(size_t index) const;

	/// Retrieve the name of the given member
	const char*getMemberName(size_t index) const;

	/// Look up the index of of a named member
	size_t getMemberIndex(const char*name) const;

	/// Return the number of members
	size_t size() const;
private:
	static const int DUT_MAXLEN = 32;
	char type_name[DUT_MAXLEN];
	char type_namespace[DUT_MAXLEN];

	bool implicit_conversions, privatetype;

	struct TypeInfo {
		char member_name[DUT_MAXLEN];
		TypeHandle handle;
	};

	// Do not use wmcore_allocator here, as this type will often be created statically in a DLL before being linked to the core. Thus we can't use the core allocator...
//	std::vector<TypeInfo, wmcore_allocator<TypeInfo>> members;
	std::vector<TypeInfo> members;
	typedef std::pair<int, UserTypeHandle> DeferredMember;
	std::vector<DeferredMember> deferred_members;
};


/**
UserTypeHandle is useful because registering a DefineUserType can only be done once your plugin has connected to the WM application core; that means that any handles
that you attempt to define statically in your cpp file will segfault as the core is not available yet.

UserTypeHandle gets around this by delaying registration of the type until first use. By the time your device constructor, etc etc are using the type handle,
the core will be connected. Please note that because of this, you must not convert (dereference) the proxy until your plugin DLL's SetupGlobalAccessPtr() function has been called to set the core ptr.

*/
class UserTypeHandle {
public:
	UserTypeHandle(DefineUserType const&type_);

	/// Conversion operator to TypeHandle. The proxy will convert its defined type to a type handle upon the first reference, caching the handle after that.
	operator PacketType();
	operator PacketTypeMany();
	TypeHandle getHandle();
	const DefineUserType &getType() const;

private:
	TypeHandle handle;
	DefineUserType type;
};

};
