#pragma once
#include "WMPacket.h"
#include "Parameter.h"

/// A Parameter packet is simply a Parameter wrapped in a WMPacket for transfer through the WM device network.
///	@ingroup data
class ParamPacket : public WMPacket
{
public:
	/// Create a blank parameter packet
	static ParamPointer CreateNew();
	/// Create a parameter packet, copying the parameter information from the constructed-from parameter
	static ParamPointer CreateNew(Parameter createfrom);

	virtual PacketPointer Clone() const;
	ParamPacket &operator=(const ParamPacket &other);

	virtual size_t getHeapUsage() { return 0;	};

	static WorldMachine::PacketType getStaticType() { return WorldMachine::Packet::parameter; };
	virtual WorldMachine::PacketType getType() const override { return getStaticType(); };
	virtual const char *getTypeName() const { return "Parameter"; };
	virtual bool queryType(WorldMachine::PacketType const&type) const;	

	virtual bool Load(std::istream &input);
	virtual bool Save(std::ostream &output);	

	virtual dword Color() { return 0x0ff0000; };

	int paramType() { return param.type; };
	VAL &paramVal() { return param.value; };	// reference to parameter value
	virtual void Clear();

	Parameter &getParam();
	void setParam(const Parameter &p);

protected:
	friend class WMPacketFactory;
	ParamPacket(void);
	~ParamPacket(void);
	virtual void resetContents() { };

	Parameter param;			// the included parameter
};


/// Allows for transferring an entire set of parameters through the device network.
/// This is useful for scenarios like:
///   transferring many parameters at once from one device to another related one
///   transferring sets of parameters from a macro to an internal device
/// @ingroup data
class ParamGroupPacket : public WMPacket 
{
public:
	static ScalarPointer CreateNew();
	ParamGroupPacket &operator=(const ParamGroupPacket &other);

	virtual size_t getHeapUsage() { return 0;	};

	virtual WorldMachine::PacketType getType() const { return WorldMachine::Packet::scalar; };
	virtual const char *getTypeName() const { return "Parameter Group"; };
	virtual void Clear() { };
	virtual PacketPointer Clone() const;

	virtual bool Load(std::istream &input);
	virtual bool Save(std::ostream &output);	

	virtual dword Color() { return 0x0770000; };

protected:
	friend class WMPacketFactory;
	ParamGroupPacket(void);
	~ParamGroupPacket(void);
	virtual void resetContents() { };

	ParameterSet params;
};


/// Wraps a single floating point value for transfer through the network.
/// Thus, it is perhaps the simplest WMPacket to be implemented.
/// @ingroup data
class ScalarPacket : public WMPacket 
{
public:
	static ScalarPointer CreateNew();
	ScalarPacket &operator=(const ScalarPacket &other);

	virtual size_t getHeapUsage() { return 0;	};

	virtual WorldMachine::PacketType getType() const { return WorldMachine::Packet::scalar; };
	virtual const char *getTypeName() const { return "Scalar"; };
	virtual void Clear() { };
	virtual PacketPointer Clone() const;

	virtual bool Load(std::istream &input);
	virtual bool Save(std::ostream &output);	

	virtual dword Color() { return 0x0770000; };

	float val() { return scalar; };
	void setval(float _newval) { scalar = _newval; };

protected:
	friend class WMPacketFactory;
	ScalarPacket(void);
	~ScalarPacket(void);
	virtual void resetContents() { };

	float scalar;
};

/** Accessor inline function to provide a safer way to cast WMPacket to ParamPacket
// in debug mode it will assert if the packet is not a ParamPacket; in release you pay nothing
// framework should never allow this anyways, but it never hurts to be careful!
	@relates ParamPacket
*/
inline ParamPointer PARAM(PacketPointer pack) { 
	if (pack) {
		bool ok = pack->getType() == WorldMachine::Packet::parameter;
		if (!ok)
			return NULL;
	}
	return ParamPointer((ParamPacket*) pack.get()); 
};

/** Accessor inline function to provide a safer way to cast WMPacket to ScalarPacket
// in debug mode it will assert if the packet is not a ScalarPacket; in release you pay nothing
// framework should never allow this anyways, but it never hurts to be careful!
	@relates ScalarPacket
*/
inline ScalarPointer SCALAR(PacketPointer pack) { 
	if (pack) {
		bool ok = pack->getType() == WorldMachine::Packet::scalar;
		if (!ok)
			return NULL;
	}
	return ScalarPointer((ScalarPacket*) pack.get()); 
};
