#pragma once
#include "common.h"
#include "Context.h"
#include "PacketTypeSystem.h"
#include <vector>

class Device;
class WMPacket;
class OutputPort;
class InputPort;
enum PSTATE;

const int PORT_NAME_LEN = 24;

namespace WorldMachine {
	/** Basic facts about a port, such as its name, datatype, any special flags, etc */
	class PortDefinition {
	public:
		enum PORTDIR { INPUT, OUTPUT };
		enum PORTFLAGS { NORMAL = 0, FLAG_OPTIONAL = 1, FLAG_VARIABLE_RESOLUTION = 2 };

		/** Port type describes the spatial types that the port accepts. The default 99% of the time one you want here is
			ACCEPT_MATCHING_ANYSPACE_DEFAULT. All ports flagged as such will be required to have matching spatial types and resolutions.

			WORLD, LOCAL, and ANY are used when a device has to have a heterogeneous set of ports, such as a routing device, or a device that places localspace items into worldspace.
			Any of these types of ports will be validated against their spatial type but will not otherwise be modified.
		*/
		enum PORTTYPE { ACCEPT_MATCHING_ANYSPACE_DEFAULT = 0, ACCEPT_ONLY_WORLDSPACE = 1, ACCEPT_ONLY_LOCALSPACE = 4, ACCEPT_ANYSPACE = 8 };

		/// @name Housekeeping functions
		//@{
		/// Set all of the definition-related info for this port
		void SetDefinition(WorldMachine::PortDefinition &pd);


		void SetName(const char *_name);
		char *GetName() { return name; };

		/// flag port with one or several of PORTFLAGS state. Replaces currently flag state
		void SetFlag(int state) { flag = state; };
		int GetFlag() const { return flag; };

		/// set or clear a single flag without disturbing other port flags
		void SetFlagBit(int statebit) { flag |= statebit; };
		void ClearFlagBit(int statebit) { flag &= ~statebit; };
		//@}

			/// @name Type-related functions
	/// See WMPacket.h for enumations
	//@{
		WorldMachine::PacketType getPrimaryType() { return packet_type.getPrimaryType(); };
		WorldMachine::PacketTypeMany& GetMultiType() { return packet_type; };
		void SetType(WorldMachine::PacketTypeMany _new) { packet_type = _new; };
		//@}

	/// @name Localspace-related functions
	//@{
	/// Check if this port potentially accepts the given spatial type. "potentially accepts" means that the spatial type COULD be accepted;
	/// but final acceptance might depend on the inputs to other ports of a device for ports whose spatial type is matching.
	/// If you need a final answer on if it accepts this type, you must use the higher-level acceptance testing function in the device world.
		bool acceptsSpatialType(PacketContext::SPACETYPE type);

		void setSpatialType(PORTTYPE t) { spatial_type = t; };
		PORTTYPE getPortSpatialType() { return spatial_type; };
	//@}

	protected:
		char name[PORT_NAME_LEN];
		int flag;				// is this port normal or special?
		PORTTYPE spatial_type;
		WorldMachine::PacketTypeMany packet_type;
	};
}

/** Mechanism through which data moves into and out of the device.

Ports are the basic wiring glue of World Machine. Each port must have a parent device, and a device contains zero or more input Port, and zero or more OutputPort.
Each port can be connected to another port (or many in the case of OutputPorts). The interconnection of these ports configures the
device network.

Each port has an intrinsic type or types (those defined in WMPacket.h) associated with it. 

	@ingroup devices
	@ingroup data
*/
class Port : public WorldMachine::PortDefinition {
public:
	Port();
	virtual ~Port(void);
	
	void SetOwner(Device* _o) { owner = _o; };
	Device *Owner() const { return owner; };

	Coord GetLoc();

	virtual Port::PORTDIR getDirection() = 0;

///@name Graph/Link related members
	virtual Port *Target() const = 0;	// The other port
	/// Return the other device that is linked through this port, or NULL if none
	virtual Device *Other() const = 0;
//@{

/// Connect this port to another
	virtual bool MakeLink(Port *other) = 0;
/// Break the connection this port has to parameter other. Includes route points.
	virtual bool BreakLink(Port *other = NULL) =0;
/// Break any routes from this port to any other ports
	virtual bool BreakRoutes() =0;
//@}

/// @name Data Type
///@{
	/// Does this port accept the given datatype?
	virtual bool acceptsDataType(WorldMachine::PacketType type);
//@}

/// Send a notification back to our owner this port is either being connected or disconnected.
	void sendNotify( bool connecting );

	virtual void SetTarget(Port *newlink, Port *oldlink = NULL) = 0;
protected:
	Device *owner;

};

/** An InputPort may be connected only to a single other (Output-)port at a time, but it may accept more than one type of input.
	@ingroup devices
*/
class InputPort : public Port {
public:
	InputPort();
	friend class OutputPort;

	virtual Port::PORTDIR getDirection() override;

	virtual bool MakeLink( Port *other ) override;
	virtual bool BreakLink( Port *other = NULL ) override;
	virtual bool BreakRoutes() override;

	/// Return the other port linked through this port
	virtual Port *Target() const override;	// The other port

	virtual OutputPort *TargetOutput() const { return target; };

	/// Return the other device that is linked through this port, or NULL if none
	virtual Device *Other() const override;

protected:
	virtual void SetTarget(Port *newlink, Port *oldlink = NULL) override;

	OutputPort *target;
};

/** An OutputPort has a single concrete type, but it may be connected to multiple other (Input-) ports
	@ingroup devices
*/
class OutputPort : public Port {
public:
	OutputPort();
	virtual ~OutputPort(void);

	virtual Port::PORTDIR getDirection() override;

	friend class InputPort;

/// Return a specific output target
	InputPort *Target(int index) const;
/// Return a specific output target's owner
	Device *Other(int index) const;

	/// Return the other port linked through this port
	virtual Port *Target() const override { return Target(0); };	// The other port
	/// Return the other device that is linked through this port, or NULL if none
	virtual Device *Other() const override { return Other(0); };

	int numLinks() { return int(outlist.size()); };
	virtual bool MakeLink(Port *other) override;
	virtual bool BreakLink(Port *other = NULL) override;
	virtual bool BreakRoutes() override;

protected:
	virtual void SetTarget(Port *newlink, Port *oldlink = NULL) override;
	std::vector<InputPort*> outlist;
};
