#pragma once
#include "ForwardDef.h"
#include "wmpacket.h"
#include "WMAllocator.h"
#include <vector>

namespace WorldMachine {

/// The DynamicTypePacket is the base type for all user-defined datatypes. World Machine can use heightfield, bitmap, curve, etc primitives.
/// User defined types are one or more primitives combined together, bound with a name. See UserPacketInterface.h for full details!
/// @ingroup data
class UserTypePacket :
	public WMPacket
{
public:
	UserTypePacket &operator=(const UserTypePacket &other);

	/// Create a user-defined packet. Note that this creates ONLY the carrying structure; all of the internal member pointers are null. you must create or assign each yourself.
	static UserTypePointer CreatePacket(UserTypeHandle &udt);

	/// Create a user-defined packet, as well as recursively create any of its members using the supplied context info.
	static UserTypePointer CreateAndInitializePacket(PacketType udt, PacketContext const&context);

	virtual size_t getHeapUsage() { return 0;	};

	virtual WorldMachine::PacketType getType() const;
	virtual bool queryType( WorldMachine::PacketType const&type ) const;
	static WorldMachine::PacketType getStaticType() { return WMP_dyntype; };

	virtual const char *getTypeName() const;
	virtual void Clear() { };
	virtual PacketPointer Clone() const;

	virtual bool Load(std::istream &input);
	virtual bool Save(std::ostream &output);	

	virtual dword Color() { return 0x03fffff; };

	//// Unique members of DTP

	/// Returns true if this type's members should never be enumerated by the user interface of the app
	virtual bool isPrivate();

	/// Return the number of defined datatypes contained within this struc
	virtual size_t numMembers();

	/// Retrieve the packet at a given index. Specifying an index greater than the number defined will return nullptr.
	/// This is intended mostly for iteration over all members; usually you will be looking up via the member(char*) method.
	virtual PacketPointer member(size_t index) const;
	/// Retrieve a member packet with the given name. Specifying an index greater than the number defined will return nullptr.
	virtual PacketPointer member(const char *name) const;

	/// Retrieve a mutable reference to the packet at a given index. This allows you to set the member
	virtual PacketPointer& member(size_t index);
	/// Retrieve a mutable reference to the packet of a given name. This allows you to set the member. If it does not exist, will throw.
	virtual PacketPointer& member(const char *name);
protected:
	UserTypePacket(TypeHandle type_handle, DefineUserType const&type_descriptor);
	~UserTypePacket(void);
	friend class WMPacketFactory;
	virtual void resetContents() { };

	virtual bool isTypePageable() { return false; };
private:
	TypeHandle udt_handle;
	DefineUserType descriptor;
	std::vector<PacketPointer, wmcore_allocator<PacketPointer>> members;
};

inline UserTypePointer USERPACKET(PacketPointer const& packet, PacketType type_required) {
	if (packet) {
		bool ok = packet->queryType(type_required);
		if (ok)
			return UserTypePointer(static_cast<UserTypePacket*>(packet.get()));
	}
	return nullptr;
};

/***********************************************************************

TODO: THIS IS BROKEN!!! Need to investigate why this doesn't work. For now, simply implement wrappers without the template impl helper.

Since we don't have reflection, we can't automatically make this. But for each built-in composite type there is a wrapper that inherits from this.
We use the CRTP to have static inheritance from this skeleton
/************************************************************************/
template < class T >
struct UserTypeWrapperSkeleton {
public:
	UserTypeWrapperSkeleton(UserTypeWrapperSkeleton const& other) {

	}

	~UserTypeWrapperSkeleton() {  };

	UserTypePointer detach() { storeMembers(); auto rval = utp; utp = nullptr; return rval; };

public:
	/// You have to define the following two functions, and all of your publicly available members.
	/// All of the publicly available members go here..

	/// Transfer members from the type pointer
	void fillMembers() { static_cast<T*>(this)->fillMembers(); };

	/// Transfer all of the data from the structure here back into the packet
	void storeMembers() { static_cast<T*>(this)->storeMembers(); };

protected:
	UserTypePointer utp;

	friend T;
	UserTypeWrapperSkeleton(PacketPointer const& pp_in) {
		utp = USERPACKET(pp_in, static_cast<T*>(this)->getType());
		if (!utp) {
			utp = UserTypePacket::CreatePacket(static_cast<T*>(this)->getType());
		};
		static_cast<T*>(this)->fillMembers();
	};
private:
	UserTypeWrapperSkeleton() { };

};


};